/* WC Clock */

/*

TODO:
- refactor to use classes :)
- support for moving clock during runtime
- custom clock hand shapes
- undrift sleep()
- systray icon and config UI
- resolution change issues?

*/

#define _WIN32_WINNT 0x0500
#define PI 3.1415926538

#include <windows.h>
#include <gdiplus.h>

#include <cmath>
#include <cstdio>

#include <vector>

using namespace std;

class Monitors {  
public:
  Monitors()
  {
    EnumDisplayMonitors(NULL, NULL, enumproc, reinterpret_cast<LPARAM>(this));
  }

  size_t count() const { return m_monitors.size(); }

  size_t primary() const {
    for (size_t i=0; i<count(); i++)
      if (m_monitors[i].dwFlags & MONITORINFOF_PRIMARY)
        return i;
    return 0;
  }

  const RECT rect(size_t i) const { return m_monitors[i].rcMonitor; }
  const RECT workrect(size_t i) const { return m_monitors[i].rcWork; }
  const char* name(size_t i) const { return m_monitors[i].szDevice; }

private:
  vector<MONITORINFOEX> m_monitors;

  void do_enum(HMONITOR mon)
  {
    MONITORINFOEX mi;
    ZeroMemory(&mi, sizeof(mi));
    mi.cbSize = sizeof(mi);
    GetMonitorInfo(mon, &mi);
    m_monitors.push_back(mi);
  }

  static BOOL CALLBACK enumproc(HMONITOR hMonitor, HDC, LPRECT, LPARAM dwData)
  { reinterpret_cast<Monitors*>(dwData)->do_enum(hMonitor); return TRUE; }
};

struct polygon {
  vector<Gdiplus::PointF> points;
};

namespace config {
  using namespace Gdiplus;

  // colors
  Color hour_fill, hour_line, hour_shadow;
  Color minute_fill, minute_line, minute_shadow;
  Color second_fill, second_line, second_shadow;
  Color hand_shadow;
  Color dot_large_fill, dot_large_line, dot_large_shadow;
  Color dot_small_fill, dot_small_line, dot_small_shadow;  

  bool merge_shadows;
  int shadow_x, shadow_y;

  // other
  bool smooth_hands;
  bool antialias;
  bool offset_workarea;
  DWORD priority;

  // sizes
  float hour_w, hour_h;
  float minute_w, minute_h;
  float second_w, second_h;
  float dot_small_r, dot_large_r;

  // shapes
  polygon hand_shape;

  // stuff
  size_t monitor;
  int clock_w, clock_h;
  int clock_x, clock_y;

  int update_interval;

  #define cfg_verify(x) if (!(x)) cfg_error(section, item)

  void cfg_error(char *section, char *item)
  {
    char text[200];
    sprintf(text, "Invalid value for %s / %s", section, item);
    MessageBox(0, text, "Error", MB_OK);
    exit(1);
  }

  int h2i(char h)
  {
    if (h>='0' && h<='9')
      return h-'0';
    if (h>='A' && h<='F')
      return h-'A'+10;
    if (h>='a' && h<='f')
      return h-'a'+10;
    return -1;
  }

  int color_helper(char *s)
  {
    int c = h2i(s[0])*16 + h2i(s[1]);
    return c;
  }

  Color read_color(char *section, char *item, char *def, char *fname)
  {
    char buf[100];
    GetPrivateProfileString(section, item, def, buf, 100, fname);
    cfg_verify(buf[0] == '#');
    cfg_verify(strlen(buf) == 9);
    int a = color_helper(buf+1);
    int r = color_helper(buf+3);
    int g = color_helper(buf+5);
    int b = color_helper(buf+7);
    cfg_verify(a>=0 && r>=0 && g>=0 && b>=0);
    return Color(a, r, g, b);
  }

  float read_float(char *section, char *item, char *def, char *fname)
  {
    char buf[100];
    GetPrivateProfileString(section, item, def, buf, 100, fname);
    float a = atof(buf);
    cfg_verify(a > 0);
    return a;
  }

  int read_int(char *section, char *item, char *def, char *fname)
  {
    char buf[100];
    GetPrivateProfileString(section, item, def, buf, 100, fname);
    int i = atoi(buf);
    return i;
  }

  bool read_bool(char *section, char *item, char *def, char *fname)
  {
    char buf[100];
    GetPrivateProfileString(section, item, def, buf, 100, fname);
    cfg_verify(!strcmpi(buf,"true") || !strcmpi(buf, "false"));
    return strcmpi(buf, "true") == 0;
  }

  DWORD read_priority(char *section, char *item, char *def, char *fname)
  {
    char buf[100];
    GetPrivateProfileString(section, item, def, buf, 100, fname);
    cfg_verify(
      !strcmpi(buf, "idle") ||
      !strcmpi(buf, "below normal") ||
      !strcmpi(buf, "normal") ||
      !strcmpi(buf, "above normal") ||
      !strcmpi(buf, "high") ||
      !strcmpi(buf, "realtime") ||
      !strcmpi(buf, "ignore")
      );

    if (0 == strcmpi(buf, "idle"))         return IDLE_PRIORITY_CLASS;
    if (0 == strcmpi(buf, "below normal")) return BELOW_NORMAL_PRIORITY_CLASS;
    if (0 == strcmpi(buf, "normal"))       return NORMAL_PRIORITY_CLASS;
    if (0 == strcmpi(buf, "above normal")) return ABOVE_NORMAL_PRIORITY_CLASS;
    if (0 == strcmpi(buf, "high"))         return HIGH_PRIORITY_CLASS;
    if (0 == strcmpi(buf, "realtime"))     return REALTIME_PRIORITY_CLASS;
    return -1;
  }

  polygon read_polygon(char *section, char *item, char *def, char *fname)
  {
    char buf[1000];
    polygon tmp;
    GetPrivateProfileString(section, item, def, buf, 1000, fname);
    int p=0;
    while(1) {
      int x, y, n=0;
      int t = sscanf(buf+p, "(%d,%d)%n", &x, &y, &n);
      cfg_verify((t==2||t==3)&&n>=5);
      p+=n;
      tmp.points.push_back(PointF(x,y));      
      for (;buf[p]==' ';p++) /* skip whitespace */ ;
      if (!buf[p])
        return tmp;
      cfg_verify(buf[p++]==',');
      for (;buf[p]==' ';p++) /* skip whitespace */ ;
    }
  }

  void monitor_help(const Monitors &m)
  {
    char buf[1024];
    sprintf(buf, "%d monitor%s detected:\n", m.count(), (m.count()==1)?"":"s");
    size_t primary = m.primary();
    for(size_t i=0; i<m.count(); i++) {
      RECT rc = m.rect(i);
      sprintf(buf+strlen(buf), "%3d: %dx%d  (%d,%d)-(%d,%d)  \"%s\"",
        i, rc.right-rc.left, rc.bottom-rc.top, rc.left, rc.top, rc.right, rc.bottom, m.name(i));
      if (i==primary)
        strcat(buf+strlen(buf), " (PRIMARY)");
    }
    MessageBox(0, buf, "WC-Clock Monitor Info", MB_OK);
    exit(1);
  }

  void load()
  {
    char fname[255];
    GetCurrentDirectory(230, fname);
    strcat(fname, "\\wc-clock.ini");

    // hand colors
    hour_fill     = read_color("colors", "hour fill",     "#FF808080", fname);
    hour_line     = read_color("colors", "hour line",     "#FF000000", fname);
    hour_shadow   = read_color("colors", "hour shadow",   "#80000000", fname);
    minute_fill   = read_color("colors", "minute fill",   "#FF808080", fname);
    minute_line   = read_color("colors", "minute line",   "#FF000000", fname);
    minute_shadow = read_color("colors", "minute shadow", "#80000000", fname);
    second_fill   = read_color("colors", "second fill",   "#FF808080", fname);
    second_line   = read_color("colors", "second line",   "#FF000000", fname);
    second_shadow = read_color("colors", "second shadow", "#80000000", fname);

    // shadow options
    hand_shadow   = read_color("colors", "hand shadow",   "#80000000", fname);
    merge_shadows = read_bool ("colors", "merge shadows", "true",      fname);

    shadow_x = read_int("colors", "shadow x", "3", fname);
    shadow_y = read_int("colors", "shadow y", "3", fname);   

    // other colors
    dot_small_fill   = read_color("colors", "small dot fill",   "#FF808080", fname);
    dot_small_line   = read_color("colors", "small dot line",   "#FF000000", fname);
    dot_small_shadow = read_color("colors", "small dot shadow", "#80000000", fname);
    dot_large_fill   = read_color("colors", "large dot fill",   "#FF808080", fname);
    dot_large_line   = read_color("colors", "large dot line",   "#FF000000", fname);
    dot_large_shadow = read_color("colors", "large dot shadow", "#80000000", fname);
   
    // sizes
    hour_w   = read_float("sizes", "hour hand width",    "1.0", fname);
    hour_h   = read_float("sizes", "hour hand height",   "0.6", fname);
    minute_w = read_float("sizes", "minute hand width",  "0.6", fname);
    minute_h = read_float("sizes", "minute hand height", "1.0", fname);
    second_w = read_float("sizes", "second hand width",  "0.2", fname);
    second_h = read_float("sizes", "second hand height", "1.0", fname);
    dot_small_r = read_float("sizes", "small dot radius", "2.0", fname);
    dot_large_r = read_float("sizes", "large dot radius", "4.0", fname);

    // shapes
    hand_shape = read_polygon("shapes", "hand shape", "(0,0), (-10, -30), (0, -100), (10, -30)", fname);

    // other options
    smooth_hands = read_bool("other", "smooth hands", "false", fname);
    antialias    = read_bool("other", "antialias", "true", fname);
    update_interval = read_int("other", "update interval", config::smooth_hands?"50":"1000", fname);

    monitor = read_int("other", "monitor", "-1", fname);
    clock_x = read_int("other", "clock x", "0", fname);
    clock_y = read_int("other", "clock y", "0", fname);
    clock_w = read_int("other", "clock width", "200", fname);
    clock_h = read_int("other", "clock height", "200", fname);
    offset_workarea = read_bool("other", "offset workarea", "true", fname);

    priority = read_priority("other", "priority", "below normal", fname);

    Monitors m;

    if (monitor == -1)
      monitor = m.primary();
    if (monitor>=m.count())
      monitor_help(m);

    RECT area = offset_workarea ? m.workrect(monitor) : m.rect(monitor);

    if (clock_x < 0) clock_x = area.right  + clock_x;
    else             clock_x = area.left   + clock_x;
    if (clock_y < 0) clock_y = area.bottom + clock_y;
    else             clock_y = area.top    + clock_y;
  }
}

// clock

static char bmpinfobuf[sizeof(BITMAPINFO)+16];
DWORD *surface = 0;
WNDCLASS fooclass;
BLENDFUNCTION blendfunc;
HDC screendc;
SIZE sz;
int redraw = 0;

HWND foo = 0;
HDC foodc = 0;

LRESULT CALLBACK wndproc(
  HWND hwnd,
  UINT uMsg,
  WPARAM wParam,
  LPARAM lParam)
{  
  switch(uMsg) {
  case WM_PAINT:
    if(surface&&redraw){
      POINT pt = {0, 0};
      ValidateRect(foo,NULL);
      UpdateLayeredWindow(
        foo,
        screendc, // device context for screen
        0,&sz, // position and size, pos not changing
        foodc, // dc for window surface
        &pt, // location of layer in dc
        0,
        &blendfunc,ULW_ALPHA);
      redraw=0;
    }
    return 0;
  case WM_QUIT:
  case WM_DESTROY:
  case WM_KEYDOWN:
    ExitProcess(0);
    return 0;
  }
  return DefWindowProc(hwnd,uMsg,wParam,lParam);
}

void pumpmessages()
{
  int i=0;
  MSG msg;
  while (PeekMessage(&msg,foo,0,0,PM_REMOVE)) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
    i++;
    // Apparently there is a bug which makes the application freeze.
    // Not knowing WTF is going on, I'm hoping this will "remedy" the situation ;)
    if (i>1000)
      return;
  }
}

void update()
{
  redraw=1;
  RedrawWindow(foo,0,0,RDW_INVALIDATE);
  UpdateWindow(foo);
  pumpmessages();
}

void init()
{
  HBITMAP bmp;
  BITMAPINFO *bmpinfo = (BITMAPINFO*)bmpinfobuf;

  fooclass.style = 0;
  fooclass.lpfnWndProc = wndproc;
  fooclass.cbClsExtra = 0;
  fooclass.cbWndExtra = 0;
  fooclass.hInstance = 0;
  fooclass.hIcon = NULL;
  fooclass.hCursor = LoadCursor(0,IDC_ARROW);
  fooclass.hbrBackground = CreateSolidBrush(RGB(0,0,0));
  fooclass.lpszMenuName = NULL;
  fooclass.lpszClassName = "WCClock";

  RegisterClass(&fooclass);

  foo = CreateWindowEx(
    WS_EX_TOPMOST|WS_EX_TOOLWINDOW|WS_EX_LAYERED|WS_EX_NOACTIVATE|WS_EX_TRANSPARENT ,
    "WCClock","WC Clock",
    WS_VISIBLE | WS_POPUP,
    config::clock_x, config::clock_y, config::clock_w, config::clock_h,
    NULL,(HMENU)0,0,NULL);
  
  blendfunc.BlendOp = AC_SRC_OVER;
  blendfunc.BlendFlags = 0;
  blendfunc.AlphaFormat = AC_SRC_ALPHA;
  blendfunc.SourceConstantAlpha = 0xff;

  screendc = GetDC(0);
  foodc = CreateCompatibleDC(screendc);

  ZeroMemory(bmpinfo,sizeof(BITMAPINFO)+16);
  bmpinfo->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmpinfo->bmiHeader.biWidth = config::clock_w;
  bmpinfo->bmiHeader.biHeight = config::clock_h;
  bmpinfo->bmiHeader.biPlanes = 1;
  bmpinfo->bmiHeader.biBitCount = 32;
  bmpinfo->bmiHeader.biCompression = BI_RGB;

  bmp = CreateDIBSection(screendc,bmpinfo,DIB_RGB_COLORS, (VOID**)&surface, NULL, 0);  
  SelectObject(foodc, bmp);

  sz.cx = config::clock_w;
  sz.cy = config::clock_h;

  if (config::priority != -1) {
    HANDLE hprocess = GetCurrentProcess();
    SetPriorityClass(hprocess, config::priority);
  }
}

void get_time(float &hour_a, float &minute_a, float &second_a)
{
  SYSTEMTIME t;
  GetLocalTime(&t);
  
  if (! config::smooth_hands) {
    hour_a = t.wHour * 360.0f/12;
    minute_a = t.wMinute * 360.0f/60;
    second_a = t.wSecond * 360.0f/60;    
  } else {
    float s = t.wMilliseconds/1000.0f + t.wSecond + t.wMinute*60 + t.wHour*60*60;
    hour_a = (s/3600.0f) * 360.0f/12;
    minute_a = (fmod(s,3600)/60.0f) * 360.0f/60;
    second_a = fmod(s,60) * 360.0f/60;
  }
}

void draw_dots(Gdiplus::Graphics &graphics, int width, int height)
{
  using namespace Gdiplus;
  // draw helper dots

  Pen dotpen_s(config::dot_small_line, 1);
  Pen dotpen_l(config::dot_large_line, 1);
  SolidBrush dotbrush_s(config::dot_small_fill);
  SolidBrush dotshadow_s(config::dot_small_shadow);
  SolidBrush dotbrush_l(config::dot_large_fill);
  SolidBrush dotshadow_l(config::dot_large_shadow);
  for(int i=0; i<60; ++i) {
    float margin = max(config::dot_small_r, config::dot_large_r)/2 + 1;
    float a = i * 2*PI / 60;
    float x = (width-margin*2)/2 + ((width-margin*2)/2)*cos(a) + margin;
    float y = (height-margin*2)/2 + ((height-margin*2)/2)*sin(a) + margin;

    float w = config::dot_small_r;
    SolidBrush *shadow(&dotshadow_s), *brush(&dotbrush_s);
    Pen *pen(&dotpen_s);

    if (i%5 == 0) {
      w = config::dot_large_r;
      shadow = &dotshadow_l;
      brush = &dotbrush_l;
      pen = &dotpen_l;
    }
    graphics.FillEllipse(shadow, x-w/2+1, y-w/2+1, w, w);
    graphics.FillEllipse(brush, x-w/2, y-w/2, w, w);
    graphics.DrawEllipse(pen, x-w/2, y-w/2, w, w);    
  }
}

void draw_clock()
{
  using namespace Gdiplus;
  Graphics graphics(foodc);
  graphics.SetSmoothingMode(config::antialias ? SmoothingModeAntiAlias : SmoothingModeNone);
  int width = sz.cx-6;
  int height = sz.cy-6;
  
  float hour_a, minute_a, second_a;
  get_time(hour_a, minute_a, second_a);

  draw_dots(graphics, width, height);

  // make paths

  GraphicsPath hand;
  hand.AddPolygon(&(config::hand_shape.points.front()), config::hand_shape.points.size());
  
  GraphicsPath *hourhand = hand.Clone();
  GraphicsPath *minutehand = hand.Clone();
  GraphicsPath *secondhand = hand.Clone();

  Matrix trans;

  // make hand paths
  
  trans.Reset();
  trans.Scale(width/200.0f * config::hour_w, height/200.0f * config::hour_h, MatrixOrderAppend);
  trans.Rotate(hour_a, MatrixOrderAppend);
  trans.Translate(width/2.0f, height/2.0f, MatrixOrderAppend);
  hourhand->Transform(&trans);

  trans.Reset();
  trans.Scale(width/200.0f * config::minute_w, height/200.0f * config::minute_h, MatrixOrderAppend);
  trans.Rotate(minute_a, MatrixOrderAppend);
  trans.Translate(width/2.0f, height/2.0f, MatrixOrderAppend);
  minutehand->Transform(&trans);

  trans.Reset();
  trans.Scale(width/200.0f * config::second_w, height/200.0f * config::second_h, MatrixOrderAppend);
  trans.Rotate(second_a, MatrixOrderAppend);
  trans.Translate(width/2.0f, height/2.0f, MatrixOrderAppend);
  secondhand->Transform(&trans);

  // draw shadow

  trans.Reset();
  trans.Translate(config::shadow_x, config::shadow_y);

  if (config::merge_shadows) {
    GraphicsPath shadowpath;
    shadowpath.AddPath(hourhand, FALSE);
    shadowpath.AddPath(minutehand, FALSE);
    shadowpath.AddPath(secondhand, FALSE);
    shadowpath.SetFillMode(FillModeWinding);
    shadowpath.Transform(&trans);

    SolidBrush shadowbrush(config::hand_shadow);
    graphics.FillPath(&shadowbrush, &shadowpath);
  } else {
    GraphicsPath *shadow_h = hourhand->Clone();
    GraphicsPath *shadow_m = minutehand->Clone();
    GraphicsPath *shadow_s = secondhand->Clone();
    shadow_h->Transform(&trans);
    shadow_m->Transform(&trans);
    shadow_s->Transform(&trans);

    SolidBrush shadowbrush_h(config::hour_shadow);
    SolidBrush shadowbrush_m(config::minute_shadow);
    SolidBrush shadowbrush_s(config::second_shadow);

    graphics.FillPath(&shadowbrush_h, shadow_h);
    graphics.FillPath(&shadowbrush_m, shadow_m);
    graphics.FillPath(&shadowbrush_s, shadow_s);

    delete shadow_h;
    delete shadow_m;
    delete shadow_s;
  }

  // draw hands

  SolidBrush hourbrush(config::hour_fill);
  Pen hourpen(config::hour_line);
  graphics.FillPath(&hourbrush, hourhand);
  graphics.DrawPath(&hourpen, hourhand);

  SolidBrush minutebrush(config::minute_fill);
  Pen minutepen(config::minute_line);
  graphics.FillPath(&minutebrush, minutehand);
  graphics.DrawPath(&minutepen, minutehand);

  SolidBrush secondbrush(config::second_fill);
  Pen secondpen(config::second_line);
  graphics.FillPath(&secondbrush, secondhand);
  graphics.DrawPath(&secondpen, secondhand);
  
  delete hourhand;
  delete minutehand;
  delete secondhand;
}

int WINAPI WinMain(HINSTANCE hinst, HINSTANCE hprev, LPSTR cmdline, int show)
{
  using namespace Gdiplus;
  GdiplusStartupInput gpsi;
  ULONG_PTR gpt;
  GdiplusStartup(&gpt, &gpsi, 0);

  config::load();

  init();

  while(1) {
    ZeroMemory(surface, sz.cx*sz.cy*sizeof(DWORD));
    draw_clock();

    update();
    Sleep(config::update_interval);
  }
 
  UnregisterClass("WCClock", NULL);
  GdiplusShutdown(gpt);

  return 0;
}
